/* \make fit
  this is version to be called from plot, cf. myfit.c
  
  getdata:
method: -1: steepest descent
         0: amoeba (default)
         1: conjugate gradient
         2: Newton-Raphson
         3: Monte Carlo 
         9: exit getdata loop
D: step to calculate numerical derivatives (method=-1,1,2 only, default=1e-5)
eps: accuracy of the minimum (irrelevant for method=3, default=1e-8)
par: `nonsphericity' valley param. par>=1 (method=3 only, default=1)
     par<=0: repeats w. extrapolation (narrow valley, all methods)
maxit: (max) number of steps 
nerr: # of samples to calculate errors (default=0=do not calculate err.)
err: err=-1: uses stdev of data provided (assuming 1 if not provided)
     err=-2: as err=-1 if stdev of data available, err=0 otherwise (default)
     err=-3: stdev of dy_i is proportional to 3rd column,
             rescaled by stdev calculated from residual sum of squares
             (for data with known weights~1/dy_i^2 but unknown precision)
     err>=0: uses stdev calc. from resid. sum sq., multiplied by y^err:
     err=0: the same abs.err. (of stdev above) of all data
     err=1: the same rel.err. of all data
     err=0.5: error is proportional to y^0.5 (suited for histogram data)
*/

#include "ground.h"
#include "minimize.h"
#include "rndgen.h"
#include "qmin.h"

// #define N (created by plot, hot key I)
#include "n.h"

REAL A[N],A0[N];

REAL func(REAL *P,REAL A)
{
  return 
// the formula (created by plot)
#include "f.c"
    ;
}

REAL Asum[N],Aqsum[N];
volatile int sig;

#define MAXn 32768
int n;
double x[MAXn],y[MAXn],dy[MAXn],dy0[MAXn],y_0[MAXn];

REAL ff(REAL *A)
{
  REAL s=0,f;
  int i;

  loop (i,0,n) {
    f=func(A,x[i]);
    s+=Sqr((f-y[i])/dy[i]); }

  return s;
}

REAL xsqrt(REAL x)
{
  if (x<0) return 0;
  else return sqrt(x);
}

void dody(double err,double sigma)
{
  int i;

  if (err==-3) loop (i,0,n) dy[i]=dy0[i]*sigma;
  else if (err<0) loop (i,0,n) dy[i]=dy0[i];
  else loop (i,0,n) dy[i]=pow(fabs(y[i]),err)*sigma;
}

int main(int narg,char **arg)
{
  char line[1024];
  FILE *f;
  REAL D=1e-5,eps=1e-8,par=1;
  int method=0,maxit=1000;
  int i,nerr=0,ierr,isstdev=0;
  double err=-2,sigma,sqsum;

  initscroll(0);
  rndinit(0,0);

  if (narg<2) {
    fprintf(stderr,"Call by (normally from plot):\n\
  fit INPUTFILE\n");
    exit(0); }

  f=fopen(arg[1],"rt");
  if (!f) ERROR(("open %s",arg[1]))

  while (fgets(line,1024,f)) {
    if (line[0]=='#') continue;
    if (n>=MAXn) ERROR(("tab ofl, increase MAXn and recompile"))
    i=sscanf(line,"%lf%lf%lf",x+n,y+n,dy0+n);
    if (i==1) ERROR(("%s bad format",line))
    if (i<=0) continue;
    if (i==2) {
      if (n) dy0[n]=dy0[n-1];
      else dy0[n]=1; }
    if (i==3) isstdev++;
    n++; }

  for (;;) {

    getdata
      getvec(A,,N)
      get(method)
      get(maxit) get(eps) get(par) get(D) 
      get(nerr) get(err)
      checkdata enddata

    if (method<-1 || method>4) return 0;

    if (err==-2 && !isstdev) err=0;

    dody(err,1);
    Minimize(method,N,A,ff,maxit,eps,D,par);

    sqsum=Val(ff(A));
    sigma=sqrt(sqsum/(n-N));

    prt("sigma=%g, adj. to deg.f.=%g",(double)(sqrt(sqsum/n)),(double)sigma);

    {
      FILE *f=fopen("fit.fit","wt");
      loop (i,0,N) fprintf(f,"# %c=%.15g\n",i+'a',Val(A[i]));
      fprintf(f,"#   x      y_data    y_fit   stdev_data  (y_fit-y_data)/stdev_data\n");
      
      loop (i,0,n)
	fprintf(f,"%.8g %.8g %.8g %.8g %.8g\n",
		x[i],y[i],Val(func(A,x[i])),dy[i],Val((func(A,x[i])-y[i])/dy[i]));
      fclose(f);
    }

    loop (i,0,N) {
      prt("%c=%.15g",i+'a',Val(A[i]));
      A0[i]=A[i];
      Asum[i]=Aqsum[i]=0; }
     
    loop (i,0,n) y_0[i]=y[i];

    if (nerr) {
      FILE *ferr=fopen("fit.err","wt");
    
      dody(err,sigma);

#if 0
      // debug
      Minimize(method,N,A,ff,maxit,eps,D,par);
      loop (i,0,N)
	prt("A[%d]=%.15g (debug)",i,A[i]);
#endif

      loop (ierr,0,nerr) {
	loop (i,0,N) A[i]=A0[i];
	loop (i,0,n) y[i]=y_0[i]+rndgauss()*dy[i];

	Minimize(method,N,A,ff,maxit,eps,D,par);
	loop (i,0,N) {
	  Asum[i]+=A[i];
	  Aqsum[i]+=Sqr(A[i]); } }
     
      loop (i,0,n) y[i]=y_0[i];

      if (nerr) prt("#     the_fit    <perturbed_fits>  stderr");
                     
      loop (i,0,N) {
	A[i]=A0[i];
	if (nerr) {
          REAL Aerr=xsqrt((Aqsum[i]/nerr-Sqr(Asum[i]/nerr))/(1.-1./nerr));
 	  prt("%c=%14.11f %14.11f %14.11f",
	         i+'a',  Val(A[i]),   Val(Asum[i]/nerr), Val(Aerr));
          fprintf(ferr," %14.11f %14.11f",
                         Val(A[i]),   Val(Aerr)); }
        Asum[i]=Aqsum[i]=0; }
      fprintf(ferr," %g\n",sigma);

    fclose(ferr); } }

  return 0;
}
