#define BUG // probably gcc bug - void code added
/* blend while diagonalizing the dipole moment crashes at command
   "loop (j,0,i) {"
   happens only with gcc -O2
   adding an empty command helps
*/

#if 0 /* test */
/* cc -O2 -o jacobins jacobins.c -lm ; jacobins
*/

#undef BUG

#include "include.h"
int sig;
#define prt printf

double Jacobins(int n,double **A,double **R,double eps);
void JacobiEigenvectors(int n,double **A0,double **A,double **R);

int main()
{
  int i,j,n=3;
  double **A0,**A,**R;

  alloc2Darrayzero(A0,n,n);
  alloc2Darrayzero(A,n,n);
  alloc2Darrayzero(R,n,n);

  A[0][0]=2;    A[0][1]=1;    A[0][2]=.5;
  A[1][0]=0.75; A[1][1]=1.5;  A[1][2]=.25;
  A[2][0]=3;    A[2][1]=-.75; A[2][2]=2;
  /*
 lambda              R[i][0]               R[i][1]          R[i][2]
 0.642079420065357  -0.394772660474081  0.0783540732249270  0.915431693656757
 1.38899927198341   -0.261647311043288 -0.298977769094411   0.917688933250884
 3.46892130795123   -0.478208136594034 -0.287526848627637  -0.829846545701988
   */

  loop (i,0,n) loop (j,0,n) A0[i][j]=A[i][j];

  printf("with(LinearAlgebra):\n");
  printf("A:=Matrix([");
  loop (i,0,n) {
    loop (j,0,n) printf("%c%.8g",j==0?'[':',',A[i][j]);
    printf("]%c",",]"[i==n-1]); }
  printf("):\nV:=Eigenvalues(A);\n");

  loop (i,0,n) put3(A[i][0],A[i][1],A[i][2]);

  Jacobins(n,A,R,-1e-9);

  printf("Eigenvalues:\n");
  loop (i,0,n) printf("%.10f ",A[i][i]);
  printf("\nOrthogonal Jacobi Eigenvectors:\n");
  loop (i,0,n) printf("[%.12g %.12g %.12g]\n",R[i][0],R[i][1],R[i][2]);

  JacobiEigenvectors(n,A0,A,R);
  printf("Eigenvectors:\n");
  loop (i,0,n) put3(R[i][0],R[i][1],R[i][2]);

  return 0;
}
#endif

/*
  generalization to a nonsymmetric diagonalizable matrix A
  [A. Ralston: The first course in numerical analysis, Chap. 10.4]
  input:
    n = rank
    A[n][n] (as **A) = matrix
    eps = error; eps>0 : silent, eps<0 : verbose
  output:
    the error is returned
    A[n][n] in the Schur form (upper diagnal matrix) is produced
    if (R!=NULL) then JacobiEigenvectors must be followed, then
     eigenvectors R[n][n] are returned:
     R[0][] is the eigenvector of eigenvalue A[0][0]
     R[1][] is the eigenvector of eigenvalue A[1][1]
     ...
  iterations are interruptable by ^C
  A (and R if used) must be allocated in advance
  the iterations are linear (unlike the original Jacobi method for
  symmetric matrices), because of numerical errors
*/

static int JacobiVerbose;

double Jacobins(int n,double **A,double **R,double eps)
/* transforms A into the Schur form by the Jacobi method */
{
  int i,j,p,q,nit=0;
  double m,x,y,cost,sint,Q,Qlim,QQ;
  double *tip,*tiq,*tpi,*tqi,*rp,*rq;
  int ifprt=eps<0;
  int activesig=n>0;
  double tt,app_aqq;

  n=abs(n);

  JacobiVerbose=eps<0;

  eps=fabs(eps);
  allocarray(tip,n);
  allocarray(tpi,n);
  allocarray(tiq,n);
  allocarray(tqi,n);
  if (R) {
    allocarray(rp,n);
    allocarray(rq,n); }

  /* max out-diag and on-diag elements; init R */
  m=x=-1;
  loop (i,0,n) {
    if (fabs(A[i][i])>x) x=fabs(A[i][i]);
    if (R) R[i][i]=1;
    loop (j,0,i) {
#ifdef BUG
      if (m==-999) put(m) // never happens
#endif
      if (R) R[i][j]=R[j][i]=0;
      if (fabs(A[i][j])>m) m=fabs(A[i][j]); } }
  if (m+x==0) return -1;

  Q=6./n;
  Qlim=1./n;
  QQ=Q;

  for (;;) {
    m*=exp(-Q); /* changing the threshold */
    nit=0;
    loop (q,0,n) loop (p,0,q) if (fabs(A[q][p])>m) {
      nit++;

      /* here p<q ; A[q][p]->0 */
      app_aqq=A[p][p]-A[q][q];
      tt=Sqr(app_aqq)+4*A[p][q]*A[q][p];
      if (tt<0) continue;
      tt=atan2(app_aqq+sqrt(Sqr(app_aqq)+4*A[p][q]*A[q][p]),2*A[p][q]);

      //      printf("%g tt=%g\n",Sqr(app_aqq)+4*A[p][q]*A[q][p],tt);
      sint=sin(tt);
      cost=cos(tt);

      loop (i,0,n) {
        tpi[i]=A[p][i];
        tqi[i]=A[q][i];
        tip[i]=A[i][p];
        tiq[i]=A[i][q]; }

      if (R) {
	copyarray(rp,R[p],n);
	copyarray(rq,R[q],n); }

      loop (i,0,p) {
	A[p][i]=cost*tpi[i]-sint*tqi[i];
	A[i][p]=cost*tip[i]-sint*tiq[i];
	A[q][i]=cost*tqi[i]+sint*tpi[i];
	A[i][q]=cost*tiq[i]+sint*tip[i]; }


      loop (i,p+1,q) {
	A[p][i]=cost*tpi[i]-sint*tqi[i];
	A[i][p]=cost*tip[i]-sint*tiq[i];
	A[q][i]=cost*tqi[i]+sint*tpi[i];
	A[i][q]=cost*tiq[i]+sint*tip[i]; }

      loop (i,q+1,n) {
	A[p][i]=cost*tpi[i]-sint*tqi[i];
	A[i][p]=cost*tip[i]-sint*tiq[i];
	A[q][i]=cost*tqi[i]+sint*tpi[i];
	A[i][q]=cost*tiq[i]+sint*tip[i]; }

      A[p][p]=tpi[p]*Sqr(cost)+tqi[q]*Sqr(sint)-(tpi[q]+tqi[p])*(cost*sint);
      A[q][q]=tpi[p]*Sqr(sint)+tqi[q]*Sqr(cost)+(tpi[q]+tqi[p])*(cost*sint);
      A[p][q]=tpi[q]*Sqr(cost)-tqi[p]*Sqr(sint)+(tpi[p]-tqi[q])*(cost*sint);
      A[q][p]=tqi[p]*Sqr(cost)-tpi[q]*Sqr(sint)+(tpi[p]-tqi[q])*(cost*sint);

      if (R) loop (i,0,n) {
	R[p][i]=rp[i]*cost-rq[i]*sint;
	R[q][i]=rp[i]*sint+rq[i]*cost; } }

    x=-1;
    loop (i,0,n) if (fabs(A[i][i])>x) x=fabs(A[i][i]);

    y=sqrt(n/(nit+0.5));
    Min(y,1.25) Max(y,0.8)
    Q*=y;
    QQ=QQ*0.875+Q*0.125;
    //    prt("%d %g %g %g %g  nit y=Q_quotient m/maxL=err Q=quotient smoothed",nit,y,m/x,Q,QQ);

    if (x<0 || m/x<eps || (Q<Qlim && m/x<sqrt(eps)) || (sig && activesig)) break;
    if (ifprt) fprintf(stderr,"Jacobi iter=%d, err=%g, Q=%g\n",nit,m/x,Q); }

  free(tqi); free(tpi);
  free(tiq); free(tip);
  if (R) { free(rp); free(rq); }

  if (sig && activesig) {
    prt("! Jacobi interrupted, err=%g",m/x);
    sig=0; }

  return m/x;
}

/* see lineq.c */
int GJlineq(int n, double **a, double *b, double *x)
{

/***
Solves a system of  n  linear eqns  AX=B  by the Gauss-Jordan elimination with a
partial selection of the pivotal element.  The function LinEq returns  -1  in
the case of singularity.
***/

  double aa,t;
  int i,j,k;

  /*** Gauss-Jordan elimination ***/
  loop (k,0,n-1) {
    aa=0.0;
    loop (i,k,n) {
      t=a[i][k];
      if (fabs(t)>fabs(aa)) { aa=t; j=i; } }
    if (aa==0.0) return -1;
    if (j!=k) {
      loop (i,k,n) {
        t=a[j][i]; a[j][i]=a[k][i]; a[k][i]=t; }
      t=b[j]; b[j]=b[k]; b[k]=t; }
    loop (i,k+1,n) {
      t=a[i][k]/aa;
      loop (j,k+1,n) a[i][j]-=t*a[k][j];
      b[i]-=t*b[k]; } }

  /*** back substitution ***/
  for (i=n-1; i>=0; i--) {
    if (a[i][i]==0.0) return -1;
    x[i]=b[i]/a[i][i];
    loop (j,0,i) b[j]-=x[i]*a[j][i]; }

  return 0;
}

void JacobiEigenvectors(int n,double **A0,double **A,double **R)
/*
 input:
   A0 = original matrix
   A  = Schur form (diagonal = eigenvectors)
   R  = orthogonal vectors of the Schur form
 returned:
   R[i][] = eigenvector coresponding to eigenvalue A[i][i]
*/
{
  double **C;
  double *P;
  double *X;
  int i,j,k,a,b;

  if (!R) return;

  alloc2Darray(C,n,n);
  allocarray(P,n);
  allocarray(X,n);

  /* R[0][] unchanged */

  loop (i,1,n) {
    loop (k,0,n) {

      P[k]=0; /* inefficient */
      loop (a,0,n) loop (b,0,n)
        P[k]+=R[k][a]*(A0[a][b]-A[i][i]*(a==b))*R[i][b];

      loop (j,0,i) {
        C[k][j]=0;
        loop (a,0,n) C[k][j]+=R[k][a]*R[j][a];
        C[k][j] *= A[i][i]-A[j][j]; } }

    GJlineq(i,C,P,X);
    loop  (j,0,i)
      loop (a,0,n) R[i][a]+=X[j]*R[j][a];
    if (JacobiVerbose && i>100 && (i%20==0)) fprintf(stderr,"JacobiEigenvectors: %d of %d finished at %s",i,n,myctime(mytime()));
  }

  free(X);
  free(P);
  free2Darray(C);
}

double JacobiTest(int N,double lambda,double **A, double *R)
/*
  returns |A.R - lambda R|/|R|, i.e., rel. error of R as eigenvector of A
  (NB: this error may be large for lambda=0)
*/
{
  int i,j;
  double *RR,s1=0,s2=0;

  allocarrayzero(RR,N);
  loop (i,0,N) loop (j,0,N) RR[i]+=A[i][j]*R[j];

  loop (i,0,N) s1+=Sqr(RR[i]);
  loop (i,0,N) RR[i]-=lambda*R[i];
  loop (i,0,N) s2+=Sqr(RR[i]);

  free(RR);

  return sqrt(s2/s1);
}

/*
   Gauss-Jordan elimination with partial pivoting
   matrix  a  is replaced by the inverse
   see also invmat.c
*/
double invmat(int N, double **a) /*********************************** invmat */
{
  double *b,*c;
  int *d;
  double det=1.0,y,w;
  int i,j,k;

  allocarray(b,N);
  allocarray(c,N);
  allocarray(d,N);

  loop (j,0,N) d[j]=j;

  loop (i,0,N) {
    k=i; y=a[i][i];
    loop (j,i+1,N) {
      w=a[i][j];
      if (fabs(w) > fabs(y)) { k=j; y=w; } }
    if ((det*=y)==0.0) goto ret;
    y=1.0/y;
    loop (j,0,N) {
      c[j]=a[j][k]; a[j][k]=a[j][i]; a[j][i]= -c[j]*y;
      b[j]=a[i][j]*=y; }
    a[i][i]=y;
    j=d[i]; d[i]=d[k]; d[k]=j;
    loop (k,0,N) if (k!=i) loop (j,0,N) if (j!=i) a[k][j]-=b[j]*c[k]; }

  loop (i,0,N) while ((k=d[i])!=i) {
    loop (j,0,N) {
      w=a[i][j]; a[i][j]=a[k][j]; a[k][j]=w; }
    j=d[i]; d[i]=d[k]; d[k]=j;
    det= -det; }

 ret:
  free(d);
  free(c);
  free(b);

  return det;
}
